#include <LiquidCrystal_I2C.h>
#include <avr/pgmspace.h> // For PROGMEM

// Define motor pins
#define STEP_PIN 2
#define DIR_PIN 3
#define DISABLE_PIN A3 // Define Disable output pin
#define AMP_INPUT A2
#define FREQ_INPUT A1

// LCD setup
LiquidCrystal_I2C lcd(0x27, 16, 2); // Adjust I2C address if necessary

// Delta time arrays for amplitudes 1 through 8 (stored in PROGMEM)
const long int deltaTimeArrays[8][16] PROGMEM = {
    {250000, 250000},                                // Amplitude 1
    {83333, 166666, 166666, 83333},                  // Amplitude 2
    {54086, 62053, 133860, 133860, 62053, 54086},    // Amplitude 3
    {40215, 43118, 51639, 115026, 115026, 51639, 43118, 40215}, // Amplitude 4
    {32047, 33447, 36921, 45167, 102416, 102416, 45167, 36921, 33447, 32047}, // Amplitude 5
    {26650, 27436, 29246, 32806, 40645, 93214, 93214, 40645, 32806, 29246, 27436, 26650}, // Amplitude 6
    {22814, 23300, 24376, 26313, 29818, 37257, 86118, 86118, 37257, 29818, 26313, 24376, 23300, 22814}, // Amplitude 7
    {19946, 20268, 20963, 22154, 24117, 27522, 34596, 80430, 80430, 34596, 27522, 24117, 22154, 20963, 20268, 19946} // Amplitude 8
};

// Working delta time array (used during runtime)
long int workingDeltaTimeArray[16];

// Direction arrays for amplitudes 1 through 8
long int directionArrays[8][16] = {
    {0, 1},                                // Amplitude 1
    {0, 0, 1, 1},                          // Amplitude 2
    {0, 0, 0, 1, 1, 1},                    // Amplitude 3
    {0, 0, 0, 0, 1, 1, 1, 1},              // Amplitude 4
    {0, 0, 0, 0, 0, 1, 1, 1, 1, 1},        // Amplitude 5
    {0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1},  // Amplitude 6
    {0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1}, // Amplitude 7
    {0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1} // Amplitude 8
};

// Logical variables
bool isQuad1_2 = true; // Track current quadrant group
bool amplitudeUpdatePending = false; // Track if an amplitude update is pending

// Motor state variables
int amplitude = 0;       // Current amplitude
int nextAmplitude = 0;   // Next amplitude after completing a full cycle
float frequency = 1.0;   // Current frequency in Hz
float previousFrequency = 1.0; // Previous frequency
unsigned long lastUpdateTime = 0; // Last time inputs and display were updated
unsigned long lastNormalUpdateTime = 0; // Time for regular LCD updates
int previousFreqInput = -1; // Previous frequency input A/D value

// Custom delay function to handle delays larger than 16383 µs
void customDelayMicroseconds(long delayTime) {
  if (delayTime > 16383) {
    long ms = delayTime / 1000;          // Millisecond part
    long us = delayTime % 1000;          // Remaining microsecond part
    delay(ms);                           // Millisecond delay
    if (us > 0) {
      delayMicroseconds(us);             // Microsecond delay
    }
  } else {
    delayMicroseconds(delayTime);        // Use delayMicroseconds for smaller delays
  }
}

// Function to scale delta times based on frequency
void updateWorkingDeltaTimes(int amplitude, float frequency) {
  float scaleFactor = 1.0 / frequency; // Scale factor based on frequency
  for (int i = 0; i < 16; i++) {
    long int originalDeltaTime = pgm_read_dword(&(deltaTimeArrays[amplitude - 1][i]));
    workingDeltaTimeArray[i] = originalDeltaTime * scaleFactor;
  }
}

// Function to pulse the step pin
void pulseStep() {
  digitalWrite(STEP_PIN, HIGH);
  delayMicroseconds(5); // Short pulse width
  digitalWrite(STEP_PIN, LOW);
}

void setup() {
  // Initialize motor pins
  pinMode(STEP_PIN, OUTPUT);
  pinMode(DIR_PIN, OUTPUT);
  pinMode(DISABLE_PIN, OUTPUT);

  // Set initial states
  digitalWrite(STEP_PIN, LOW);
  digitalWrite(DIR_PIN, LOW);
  digitalWrite(DISABLE_PIN, HIGH); // Motor disabled initially

  // Initialize LCD
  lcd.init();
  lcd.backlight();
  lcd.setCursor(0, 0);
  lcd.print("Amp: 0");

  // Initialize Serial communication
  Serial.begin(9600);

  // Initialize working delta times
  updateWorkingDeltaTimes(1, frequency);
}

void loop() {
  unsigned long currentTime = millis();

  // Read frequency input
  int currentFreqInput = analogRead(FREQ_INPUT);

  // Trigger immediate update if frequency input changes significantly
  if (abs(currentFreqInput - previousFreqInput) > 2) {
    lastUpdateTime = currentTime; // Reset last update time to trigger update
  }

  // Update LCD every 6 seconds normally or immediately on significant input change
  if (currentTime - lastUpdateTime >= 6000 || abs(currentFreqInput - previousFreqInput) > 2) {
    lastUpdateTime = currentTime;
    previousFreqInput = currentFreqInput;

    // Read amplitude from analog input
    int analogValueAmp = analogRead(AMP_INPUT);
    nextAmplitude = map(analogValueAmp, 0, 1023, 0, 8);

    // Read frequency from analog input
    frequency = map(currentFreqInput, 0, 1023, 10, 500) / 10.0; // Map to 1.0 to 30.0 Hz

    // If frequency or amplitude changes, update working delta times
    if (frequency != previousFrequency || nextAmplitude != amplitude) {
      if (nextAmplitude > 0) {
        updateWorkingDeltaTimes(nextAmplitude, frequency);
      }
      previousFrequency = frequency;
    }

    // Update LCD display
    lcd.setCursor(0, 0);
    lcd.print("Amp: ");
    lcd.print(nextAmplitude);
    lcd.setCursor(0, 1);
    lcd.print("Freq: ");
    lcd.print(frequency, 1);
    lcd.print(" Hz");

    // If amplitude is 0, disable the motor
    if (nextAmplitude == 0 && amplitude != 0) {
      digitalWrite(DISABLE_PIN, HIGH); // Disable motor
      amplitude = 0; // Stop driving the motor
      return;
    }

    // If motor is disabled and next amplitude is greater than 0, enable the motor
    if (amplitude == 0 && nextAmplitude > 0) {
      digitalWrite(DISABLE_PIN, LOW); // Enable motor
      amplitude = nextAmplitude; // Set amplitude to start driving
    }

    // Mark amplitude update as pending
    if (nextAmplitude != amplitude && nextAmplitude > 0) {
      amplitudeUpdatePending = true;
    }
  }

  // Skip motor driving if amplitude is 0
  if (amplitude == 0) {
    return;
  }

  // Select the correct arrays based on current amplitude
  long int* deltaTimes = workingDeltaTimeArray;
  long int* directions = directionArrays[amplitude - 1];
  int numSteps = amplitude * 2; // Total steps for the given amplitude

  // Drive the motor
  for (int i = 0; i < numSteps; i++) {
    // Determine direction based on quadrant group
    if (isQuad1_2) {
      digitalWrite(DIR_PIN, directions[i]); // Use original direction for Quadrants 1/2
    } else {
      digitalWrite(DIR_PIN, !directions[i]); // Invert direction for Quadrants 3/4
    }

    // Use the custom delay function
    customDelayMicroseconds(deltaTimes[i]);

    // Pulse the step pin
    pulseStep();
  }

  // Toggle the quadrant group after completing the cycle
  isQuad1_2 = !isQuad1_2;

  // If a full cycle is complete (Quadrants 1, 2, 3, and 4), update the amplitude
  if (!isQuad1_2 && amplitudeUpdatePending) {
    amplitude = nextAmplitude;
    amplitudeUpdatePending = false;
  }
}
  
